"""Testing dav cli."""
import os
import sys
import textwrap
from argparse import Namespace
from datetime import datetime, timedelta
from io import StringIO
from typing import Any, Dict, List
from unittest import mock

import colorama
import pytest
from fsspec.implementations.memory import MemoryFileSystem as _MemoryFS
from pytest import CaptureFixture, MonkeyPatch

from webdav4.cli import (
    Command,
    CommandCat,
    CommandCopy,
    CommandDiskUsage,
    CommandLS,
    CommandMkdir,
    CommandMove,
    CommandRemove,
    CommandRun,
    CommandSync,
    File,
    LSTheme,
    Row,
    Size,
    color_file,
    format_datetime,
    get_parser,
    human_size,
    main,
)

from .utils import TmpDir


class MemoryFileSystem(_MemoryFS):
    """Overriding to make memory per instance and extend protocol support."""

    cachable = False

    def __init__(self, *args: Any, **storage_options: Any) -> None:
        """Initializing with a custom store for each instance."""
        super().__init__(*args, **storage_options)
        self.store: Dict[str, Any] = {}
        self.pseudo_dirs: List[str] = []

    def mkdir(
        self, path: str, create_parents: bool = True, **kwargs: Any
    ) -> None:
        """Adds support for protocol."""
        path = self._strip_protocol(path)
        if create_parents and self.isdir(path):
            return
        super().mkdir(path, create_parents=create_parents, **kwargs)

    def copy(
        self,
        path1: str,
        path2: str,
        recursive: bool = False,
        on_error: str = None,
        **kwargs: Any
    ) -> None:
        """Adds support for protocol."""
        path1 = self._strip_protocol(path1)
        path2 = self._strip_protocol(path2)
        super().copy(
            path1, path2, recursive=recursive, on_error=on_error, **kwargs
        )

    def mv(self, path1: str, path2: str, **kwargs: Any) -> None:
        """Adds support for protocol."""
        path1 = self._strip_protocol(path1)
        path2 = self._strip_protocol(path2)
        super().mv(path1, path2, **kwargs)

    def rm(
        self, path: str, recursive: bool = False, maxdepth: int = None
    ) -> None:
        """Adds support for protocol."""
        path = self._strip_protocol(path)
        super().rm(path, recursive=recursive, maxdepth=maxdepth)


@pytest.fixture(autouse=True)
def unset_envvars(monkeypatch: MonkeyPatch):
    """Unset some envvars that might affect auth or coloring."""
    envvars = [
        "WEBDAV_ENDPOINT_URL",
        "LS_COLORS",
        "FORCE_COLOR",
        "NO_COLOR",
        "TERM",
    ]
    for envvar in envvars:
        monkeypatch.delenv(envvar, raising=False)


@pytest.mark.parametrize(
    "file_name, expected_color",
    [
        ("src/README.md", colorama.Fore.YELLOW),
        ("src/file.py", None),
        ("src/photo.jpg", colorama.Fore.CYAN),
        ("src/Makefile", colorama.Fore.YELLOW),
        ("src/setup.py~", colorama.Fore.LIGHTWHITE_EX),
        ("src/#temp#", colorama.Fore.LIGHTWHITE_EX),
        ("src/file.bak", colorama.Fore.LIGHTWHITE_EX),
        ("src/music.aac", colorama.Fore.RED),
        ("src/video.mov", colorama.Fore.MAGENTA),
        ("gpg/my.gpg", colorama.Fore.LIGHTGREEN_EX),
        ("my-doc.xlsx", colorama.Fore.BLUE),
        ("folder.zip", colorama.Fore.LIGHTRED_EX),
        ("a.o", colorama.Fore.GREEN),
        ("no-extension", None),
        ("no-extension.", None),
        (".no-extension", None),
        ("src/file", None),
    ],
)
def test_file_colors(file_name, expected_color):
    """Test that file are wrapped with colors based on type/extension."""
    if expected_color:
        expected = expected_color + file_name + colorama.Style.RESET_ALL
    else:
        expected = file_name
    assert color_file(File.from_path(file_name)) == expected


@pytest.mark.parametrize("file_name", ["src/file.py", "src", "src/docs"])
def test_dir_colors(file_name):
    """Test that dirs are brightly colored."""
    assert (
        color_file(File.from_path(file_name), isdir=True)
        == colorama.Style.BRIGHT
        + colorama.Fore.BLUE
        + file_name
        + colorama.Style.RESET_ALL
    )


@mock.patch.object(sys.stdout, "isatty", return_value=True)
def test_ls_colors(m):
    """Test ls colors theme with support for lscolors envvar."""
    theme = LSTheme(lscolors="rs=0:di=01;34:ex=01;32:*.py=33:*.zip=35")
    assert theme.codes == {"rs": "0", "di": "01;34", "ex": "01;32"}
    assert theme.extensions == {"py": "33", "zip": "35"}
    assert theme.colored


def test_lstheme_should_color(monkeypatch: MonkeyPatch):
    """Test LSTheme color deciding factors."""
    theme = LSTheme()
    with mock.patch.object(sys.stdout, "isatty", return_value=False):
        assert not theme._should_color()

        monkeypatch.setenv("FORCE_COLOR", "false")
        assert not theme._should_color()

        monkeypatch.setenv("FORCE_COLOR", "")
        assert not theme._should_color()

        monkeypatch.setenv("FORCE_COLOR", "0")
        assert not theme._should_color()

        monkeypatch.setenv("FORCE_COLOR", "true")
        assert theme._should_color()

        monkeypatch.setenv("FORCE_COLOR", "1")
        assert theme._should_color()

        monkeypatch.setenv("FORCE_COLOR", "2")
        assert theme._should_color()

        monkeypatch.setenv("FORCE_COLOR", "3")
        assert theme._should_color()

    monkeypatch.delenv("FORCE_COLOR")

    with mock.patch.object(sys.stdout, "isatty", return_value=True):
        assert theme._should_color()

        monkeypatch.setenv("TERM", "dumb")
        assert not theme._should_color()
        monkeypatch.delenv("TERM")

        # NO_COLOR only cares about it being set
        monkeypatch.setenv("NO_COLOR", "1")
        assert not theme._should_color()
        monkeypatch.delenv("NO_COLOR")

        monkeypatch.setenv("FORCE_COLOR", "0")
        assert not theme._should_color()

        monkeypatch.setenv("FORCE_COLOR", "1")
        assert theme._should_color()

        # should still print in color as FORCE_COLOR is set.
        monkeypatch.setenv("NO_COLOR", "1")
        assert theme._should_color()


def test_style_path():
    """Test styling path for the ls output."""
    theme = LSTheme(lscolors="*.py=32")

    theme.colored = False
    assert theme.style_path("hello") == "hello"
    assert theme.style_path("hello", isdir=True) == "hello"

    theme.colored = True

    assert theme.style_path("path") == "path"
    assert theme.extensions == {"py": "32"}
    assert theme.style_path("path.py") == "\x1b[32mpath.py\x1b[0m"
    assert theme.style_path("path", isdir=True) == "\x1b[1m\x1b[34mpath\x1b[0m"

    theme = LSTheme(dir_trailing_slash=True)
    theme.colored = True
    assert (
        theme.style_path("path", isdir=True) == "\x1b[1m\x1b[34mpath/\x1b[0m"
    )
    theme.colored = False
    assert theme.style_path("path", isdir=True) == "path/"


def test_style_size_datetime():
    """Test styling of size and datetime in the output of ls."""
    theme = LSTheme()
    theme.colored = False

    dt = datetime.today()
    assert theme.style_datetime(format_datetime(dt)) == dt.strftime(
        "%b %d %H:%M"
    )
    assert theme.style_size(*human_size(2027)) == "2.0k"

    theme.colored = True
    assert theme.style_datetime(
        format_datetime(dt)
    ) == "\x1b[34m{0}\x1b[0m".format(dt.strftime("%b %d %H:%M"))
    assert theme.style_size(
        *human_size(2027)
    ) == "\x1b[1m\x1b[32m{0}\x1b[0m\x1b[32m{1}\x1b[0m".format("2.0", "k")

    assert theme.style_size(*human_size(None)) == "\x1b[2m-\x1b[0m"


@pytest.mark.parametrize(
    "nbytes, expected",
    [
        (0, ("0", "")),
        (2027, ("2.0", "k")),
        (389, ("389", "")),
        (6498656, ("6.2", "M")),
        (1073741824, ("1.0", "G")),
        (107374182400, ("100", "G")),
        (None, ("", "-")),
    ],
)
def test_human_size(nbytes: int, expected: str):
    """Test humanizing of n bytes/sizes."""
    assert human_size(nbytes) == expected


def test_datetime_format():
    """Test humanizing datetime format that is compatible with ls."""
    today = datetime.today()
    recent = today - timedelta(days=10)
    old = today - timedelta(days=200)

    assert format_datetime(recent) == recent.strftime("%b %d %H:%M")
    assert format_datetime(old) == old.strftime("%b %d %Y")
    assert format_datetime(None) == "-"


@pytest.mark.parametrize(
    "path, text",
    [
        ("data", "6 (6 bytes) in 2 files: data"),
        ("data/foo", "3 (3 bytes) in 1 file: data/foo"),
        ("data/bar", "3 (3 bytes) in 1 file: data/bar"),
    ],
)
def test_du_cli(capsys: CaptureFixture, path: str, text: str):
    """Test du command."""
    mfs = MemoryFileSystem()
    mfs.mkdir("data")
    mfs.pipe({"data/foo": b"foo", "data/bar": b"bar"})

    ns = Namespace(path=path)
    CommandDiskUsage(ns, mfs).run()

    out, _ = capsys.readouterr()
    assert text in out
    mfs.clear_instance_cache()


def test_cat_cli(capsys: CaptureFixture):
    """Test cat command."""
    mfs = MemoryFileSystem()
    mfs.pipe({"data": b"data" * 1000})

    ns = Namespace(path="data")
    CommandCat(ns, mfs).run()
    out, _ = capsys.readouterr()
    # printed as str
    assert "data" * 1000 in out


def test_cp_cli(storage_dir: TmpDir, monkeypatch: MonkeyPatch):
    """Test cp command."""
    memfs = MemoryFileSystem()

    memfs.mkdir("data1")
    memfs.pipe({"data1/foo": b"foo", "data1/bar": b"bar"})

    ns = Namespace(
        path1="memory://data1", path2="memory://data2", recursive=False
    )
    CommandCopy(ns, memfs).run()

    assert memfs.cat("data1", recursive=True, on_error="ignore") == {
        "/data1/bar": b"bar",
        "/data1/foo": b"foo",
    }
    assert memfs.isdir("data2")

    ns = Namespace(
        path1="memory://data1", path2="memory://data2", recursive=True
    )
    CommandCopy(ns, memfs).run()
    assert memfs.cat("data1", recursive=True, on_error="ignore") == {
        "/data1/bar": b"bar",
        "/data1/foo": b"foo",
    }
    assert memfs.cat("data2", recursive=True, on_error="ignore") == {
        "/data2/bar": b"bar",
        "/data2/foo": b"foo",
    }
    assert memfs.isdir("data2")

    ns = Namespace(
        path1="memory://data1/foo",
        path2="memory://data2/foobar",
        recursive=True,
    )
    CommandCopy(ns, memfs).run()
    assert memfs.cat_file("data2/foobar") == b"foo"

    monkeypatch.chdir(storage_dir)

    d = {"lorem": "lorem", "ipsum": "ipsum"}
    storage_dir.gen({"dir": d})
    ns = Namespace(path1="dir", path2="memory://dir", recursive=True)
    CommandCopy(ns, memfs).run()
    assert memfs.cat("dir", True, "ignore") == {
        "/dir/ipsum": b"ipsum",
        "/dir/lorem": b"lorem",
    }

    ns = Namespace(path1="memory://dir", path2="dir2", recursive=True)
    CommandCopy(ns, memfs).run()
    assert storage_dir.cat() == {"dir": d, "dir2": d}

    ns = Namespace(path1="memory://dir/ipsum", path2="ipsum", recursive=True)
    CommandCopy(ns, memfs).run()
    assert storage_dir.cat() == {"dir": d, "dir2": d, "ipsum": "ipsum"}


def test_rm_cli():
    """Test rm command."""
    memfs = MemoryFileSystem()

    memfs.mkdir("data1")
    memfs.pipe({"data1/foo": b"foo", "data1/bar": b"bar"})

    cmd = CommandRemove(Namespace(path="data1", recursive=False), memfs)
    with pytest.raises(OSError):
        cmd.run()

    cmd = CommandRemove(Namespace(path="data1/bar", recursive=False), memfs)
    cmd.run()
    assert not memfs.exists("data1/bar")

    cmd = CommandRemove(Namespace(path="data1", recursive=True), memfs)
    cmd.run()
    assert not memfs.exists("data1")


def test_mkdir_cli():
    """Test mkdir command."""
    memfs = MemoryFileSystem()

    cmd = CommandMkdir(Namespace(path="data1", parents=False), memfs)
    cmd.run()
    assert memfs.isdir("data1")

    cmd = CommandMkdir(Namespace(path="data1/dir1/dir2", parents=True), memfs)
    cmd.run()
    assert memfs.isdir("data1/dir1")
    assert memfs.isdir("data1/dir1/dir2")


def test_mv_cli(storage_dir: TmpDir, monkeypatch: MonkeyPatch):
    """Test mv command."""
    memfs = MemoryFileSystem()

    memfs.mkdir("data1")
    memfs.pipe({"data1/foo": b"foo", "data1/bar": b"bar"})

    ns = Namespace(
        path1="memory://data1", path2="memory://data2", recursive=True
    )
    CommandMove(ns, memfs).run()
    assert not memfs.isdir("data1")
    assert memfs.cat("data2", recursive=True, on_error="ignore") == {
        "/data2/bar": b"bar",
        "/data2/foo": b"foo",
    }
    assert memfs.isdir("data2")

    ns = Namespace(
        path1="memory://data2/foo",
        path2="memory://data2/foobar",
        recursive=True,
    )
    CommandMove(ns, memfs).run()
    assert memfs.cat_file("data2/foobar") == b"foo"
    assert not memfs.exists("memory://data2/foo")

    monkeypatch.chdir(storage_dir)

    d = {"lorem": "lorem", "ipsum": "ipsum"}
    storage_dir.gen({"dir": d})
    ns = Namespace(path1="dir", path2="memory://dir", recursive=True)
    CommandMove(ns, memfs).run()
    assert memfs.cat("dir", True, "ignore") == {
        "/dir/ipsum": b"ipsum",
        "/dir/lorem": b"lorem",
    }

    ns = Namespace(path1="memory://dir", path2="dir2", recursive=True)
    CommandMove(ns, memfs).run()
    assert storage_dir.cat() == {"dir": d, "dir2": d}


def test_ls_cli(capsys: CaptureFixture):
    """Test ls command."""
    mfs = MemoryFileSystem()

    mfs.mkdir("data")
    mfs.mkdir("dir")
    mfs.pipe(
        {
            "data/foo": b"foo",
            "data/bar": b"bar",
            "dir/nested/foo": b"dirfoo" * 1000,
        }
    )

    ns = Namespace(path="data", recursive=False, level=None, full_path=False)
    assert set(CommandLS(ns, mfs).ls()) == {
        Row("-", Size("3"), "foo"),
        Row("-", Size("3"), "bar"),
    }

    ns = Namespace(
        path="data/foo", recursive=False, level=None, full_path=False
    )
    assert set(CommandLS(ns, mfs).ls()) == {Row("-", Size("3"), "foo")}

    ns = Namespace(
        path="data/bar", recursive=False, level=None, full_path=False
    )
    assert set(CommandLS(ns, mfs).ls()) == {Row("-", Size("3"), "bar")}

    with pytest.raises(FileNotFoundError):
        ns = Namespace(
            path="not-existing", recursive=False, level=None, full_path=False
        )
        assert set(CommandLS(ns, mfs).ls())

    ns = Namespace(path="data", recursive=False, level=None, full_path=True)
    assert set(CommandLS(ns, mfs).ls()) == {
        Row("-", Size("3"), "data/foo"),
        Row("-", Size("3"), "data/bar"),
    }

    ns = Namespace(path="", recursive=True, level=2, full_path=True)
    assert set(CommandLS(ns, mfs).ls()) == {
        Row("-", Size("3"), "data/foo"),
        Row("-", Size("3"), "data/bar"),
        Row("-", Size("0"), "dir/nested", isdir=True),
    }

    ns = Namespace(path="", recursive=True, level=None, full_path=False)
    assert set(CommandLS(ns, mfs).ls()) == {
        Row("-", Size("3"), "data/foo"),
        Row("-", Size("3"), "data/bar"),
        Row("-", Size("5.9", "k"), "dir/nested/foo"),
    }

    CommandLS.render([])
    assert ("", "") == capsys.readouterr()

    CommandLS.render(
        [
            Row("Apr 05 09:40", Size("148.0", "k"), "README.md"),
            Row("Jun 03 03:33", Size("6.2", "M"), "my-docs.docx"),
        ]
    )

    out, _ = capsys.readouterr()
    assert (
        textwrap.dedent(
            """\
        Apr 05 09:40 148.0k README.md
        Jun 03 03:33   6.2M my-docs.docx"""
        )
        in out
    )

    ns = Namespace(
        path="data/foo", recursive=True, level=None, full_path=False
    )
    CommandLS(ns, mfs).run()
    assert capsys.readouterr() == ("-  3 foo\n", "")

    with pytest.raises(FileNotFoundError):
        ns = Namespace(
            path="not-existing", recursive=True, level=None, full_path=False
        )
        CommandLS(ns, mfs).run()


def test_sync_cli_local_to_remote(
    storage_dir: TmpDir, monkeypatch: MonkeyPatch
):
    """Test syncing between local to remote filesystem."""
    storage_dir.gen({"data": {"foo": "foo"}})
    memfs = MemoryFileSystem()

    monkeypatch.chdir(storage_dir)
    ns = Namespace(path1="data", path2="memory://data", delete=False)

    cmd = CommandSync(ns, memfs)
    cmd.run()
    assert memfs.cat("data", True, "ignore") == {"/data/foo": b"foo"}

    (storage_dir / "data" / "bar").write_text("bar")
    cmd.run()
    assert memfs.cat("data", True, "ignore") == {
        "/data/foo": b"foo",
        "/data/bar": b"bar",
    }

    (storage_dir / "data" / "bar").write_text("new updates")
    cmd.run()
    assert memfs.cat("data", True, "ignore") == {
        "/data/foo": b"foo",
        "/data/bar": b"new updates",
    }

    (storage_dir / "data" / "bar").unlink()
    ns = Namespace(path1="data", path2="memory://data", delete=True)
    cmd = CommandSync(ns, memfs)
    cmd.run()
    assert memfs.cat("data", True, "ignore") == {"/data/foo": b"foo"}

    (storage_dir / "data" / "dir").gen({"lorem": "lorem"})
    cmd.run()
    assert memfs.cat("data", True, "ignore") == {
        "/data/foo": b"foo",
        "/data/dir/lorem": b"lorem",
    }

    # nothing should change
    cmd.run()
    assert memfs.cat("data", True, "ignore") == {
        "/data/foo": b"foo",
        "/data/dir/lorem": b"lorem",
    }

    # trying to make src to be a directory, we should abort in this case
    (storage_dir / "data" / "foo").unlink()
    (storage_dir / "data").gen({"foo": {"foo": "foo"}})
    with pytest.raises(TypeError) as exc_info:
        cmd.run()
    assert (
        str(exc_info.value) == "cannot sync between different types, "
        "src is directory, dest is file"
    )


def test_sync_remote_to_local(storage_dir: TmpDir, monkeypatch: MonkeyPatch):
    """Test syncing between remote to local filesystem."""
    memfs = MemoryFileSystem()
    memfs.mkdir("data")
    memfs.pipe({"data/foo": b"foo"})

    monkeypatch.chdir(storage_dir)
    ns = Namespace(path1="memory://data", path2="data", delete=False)

    cmd = CommandSync(ns, memfs)
    cmd.run()
    assert storage_dir.cat() == {"data": {"foo": "foo"}}

    memfs.pipe({"data/bar": b"bar"})
    cmd.run()
    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "bar"}}

    memfs.pipe({"data/bar": b"new updates"})
    cmd.run()
    assert storage_dir.cat() == {"data": {"foo": "foo", "bar": "new updates"}}

    memfs.rm("data/bar")
    ns = Namespace(path1="memory://data", path2="data", delete=True)
    cmd = CommandSync(ns, memfs)
    cmd.run()
    assert storage_dir.cat() == {"data": {"foo": "foo"}}

    memfs.mkdir("data/dir")
    memfs.pipe({"data/dir/lorem": b"lorem"})
    cmd.run()
    assert storage_dir.cat() == {
        "data": {"foo": "foo", "dir": {"lorem": "lorem"}}
    }

    # nothing should change
    cmd.run()
    assert storage_dir.cat() == {
        "data": {"foo": "foo", "dir": {"lorem": "lorem"}}
    }


def test_sync_remote_to_remote():
    """Test syncing between remote to remote filesystem."""
    memfs = MemoryFileSystem()

    memfs.mkdir("data")
    memfs.pipe({"data/foo": b"foo"})

    ns = Namespace(path1="memory://data", path2="memory://data2", delete=False)

    cmd = CommandSync(ns, memfs)
    cmd.run()
    assert memfs.cat("data2", True, "ignore") == {"/data2/foo": b"foo"}

    memfs.pipe({"data/bar": b"bar"})
    cmd.run()
    assert memfs.cat("data2", True, "ignore") == {
        "/data2/foo": b"foo",
        "/data2/bar": b"bar",
    }

    memfs.pipe({"data/bar": b"new updates"})
    cmd.run()
    assert memfs.cat("data2", True, "ignore") == {
        "/data2/foo": b"foo",
        "/data2/bar": b"new updates",
    }

    memfs.rm("data/bar")
    ns = Namespace(path1="memory://data", path2="memory://data2", delete=True)
    cmd = CommandSync(ns, memfs)
    cmd.run()
    assert memfs.cat("data2", True, "ignore") == {"/data2/foo": b"foo"}

    memfs.mkdir("data/dir")
    memfs.pipe({"data/dir/lorem": b"lorem"})
    cmd.run()
    assert memfs.cat("data2", True, "ignore") == {
        "/data2/foo": b"foo",
        "/data2/dir/lorem": b"lorem",
    }

    # nothing should change
    cmd.run()
    assert memfs.cat("data2", True, "ignore") == {
        "/data2/foo": b"foo",
        "/data2/dir/lorem": b"lorem",
    }


def test_run_cli(monkeypatch: MonkeyPatch):
    """Test run command."""
    memfs = MemoryFileSystem()

    memfs.mkdir("data")
    memfs.pipe({"data/foo": b"foo"})

    buff = StringIO("ls" + os.linesep)
    buff.write("du" + os.linesep)
    buff.write("sync memory://data memory://data2" + os.linesep)
    buff.write("# comments" + os.linesep)
    buff.write("rm data2 --recursive" + os.linesep)
    buff.write("cat memory://data/foo" + os.linesep)
    buff.write("mv memory://data memory://data3 --recursive" + os.linesep)
    buff.write("ls memory://data3" + os.linesep)
    buff.write("du memory://data3" + os.linesep)
    buff.write("mkdir memory://data3 -p" + os.linesep)
    buff.seek(0)

    monkeypatch.setattr("sys.stdin", buff)
    _, subparsers = get_parser()
    cmd = CommandRun(Namespace(path="-", subparsers=subparsers), memfs)
    cmd.run()

    pos = buff.tell()
    buff.write("not-existing-command" + os.linesep)
    buff.seek(pos)
    with pytest.raises(ValueError) as exc_info:
        cmd.run()
    assert str(exc_info.value) == "unknown command: not-existing-command"

    with pytest.raises(ValueError) as exc_info:
        cmd = CommandRun(Namespace(path=None, subparsers=subparsers), memfs)
        cmd.run()
    assert str(exc_info.value) == "no path specified or contents piped"


def test_shorthand_url():
    """Test curl-like shorthand url and auto adding scheme support."""
    ns = Namespace(endpoint_url=":3000/foo", user=None, password=None)
    assert Command(ns).fs.client.base_url == "http://localhost:3000/foo"

    ns.endpoint_url = ":/foo"
    assert Command(ns).fs.client.base_url == "http://localhost/foo"

    ns.endpoint_url = "example.org/path"
    assert Command(ns).fs.client.base_url == "https://example.org/path"

    ns.endpoint_url = "https://example.org/path"
    assert Command(ns).fs.client.base_url == "https://example.org/path"


def test_auth(monkeypatch: MonkeyPatch):
    """Test auth from args and url."""
    ns = Namespace(
        endpoint_url="http://server.com", user="user", password="pwd"
    )
    assert Command(ns).auth == ("user", "pwd")

    ns.endpoint_url = "http://user:pwd@server.com"
    assert Command(ns).auth == ("user", "pwd")

    monkeypatch.setenv("WEBDAV_ENDPOINT_URL", "http://user2:pwd2@server.com")
    # --endpoint-url should still take precedence
    assert Command(ns).auth == ("user", "pwd")

    ns = Namespace(endpoint_url=None, user="user", password="pwd")
    assert Command(ns).auth == ("user2", "pwd2")

    monkeypatch.delenv("WEBDAV_ENDPOINT_URL")


def test_main():
    """Test main command line entrypoint."""
    assert main(["-v", "ls"]) == 1
    assert main(["ls"]) == 1



"""Test retrying in the client."""
from http import HTTPStatus
from unittest import mock

import pytest
from httpx import Request, Response

from webdav4.client import BadGatewayError, HTTPError, ResourceLocked
from webdav4.retry import retry

request = Request("get", "example.com")


@pytest.mark.parametrize(
    "exc",
    [
        BadGatewayError(),
        HTTPError(
            Response(HTTPStatus.TOO_MANY_REQUESTS.value, request=request)
        ),
        ResourceLocked("beep boop bop"),
    ],
)
def test_retry(exc):
    """Test that the function is rerun/retried on failure."""
    response = Response(HTTPStatus.OK.value)
    func = mock.MagicMock(side_effect=[exc, response])
    assert retry(True)(func) == response
    assert func.call_count == 2


@pytest.mark.parametrize(
    "status_code, retries ",
    [
        # this is not handled, so it should not retry
        (HTTPStatus.CONFLICT, 1),
        # it tries to retry in case of this error
        (HTTPStatus.SERVICE_UNAVAILABLE, 3),
    ],
)
def test_retry_negative(status_code, retries):
    """Test retry but with the end-result being always a failure.

    Example: one case where it is asked not to be handled and the other case
    where even though the error is asked to be handled, every attempt fails
    and reaches maximum retries allowed.
    """
    func = mock.MagicMock(
        side_effect=HTTPError(Response(status_code.value, request=request))
    )
    with pytest.raises(HTTPError) as exc_info:
        retry(True)(func)
    assert exc_info.value.status_code == status_code
    assert func.call_count == retries



"""Testing multistatus responses."""
from datetime import datetime
from typing import Optional, Tuple
from xml.etree.ElementTree import Element, fromstring

import pytest
from dateutil.tz import tzutc
from httpx import Response as HTTPResponse

from webdav4.multistatus import (
    DAVProperties,
    MultiStatusResponse,
    MultiStatusResponseError,
    Response,
    parse_multistatus_response,
    prepare_propfind_request_data,
    prop,
)
from webdav4.urls import URL


def test_prop():
    """Test getting text from xml element."""
    status_line = "<d:status>HTTP/1.1 200 OK</d:status>"
    content = f'<d:multistatus xmlns:d="DAV:">{status_line}</d:multistatus>'
    elem = fromstring(content)
    assert prop(elem, "status") == "HTTP/1.1 200 OK"


def test_prop_relative():
    """Test getting text anywhere from xml element."""
    status_line = "<d:status>HTTP/1.1 200 OK</d:status>"
    response = f"<d:response>{status_line}</d:response>"
    content = f'<d:multistatus xmlns:d="DAV:">{response}</d:multistatus>'
    elem = fromstring(content)
    assert prop(elem, "status") is None
    assert prop(elem, "status", relative=True) == "HTTP/1.1 200 OK"


@pytest.mark.parametrize(
    "args",
    [
        (fromstring('<d:propstat xmlns:d="DAV:"></d:propstat>')),
        (),
    ],
)
def test_dav_properties_empty(args: Tuple[Element]):
    """Test dav properties when it's empty."""
    props = DAVProperties(*args)
    assert (
        props.raw
        == {
            "modified": None,
            "created": None,
            "content_length": None,
            "etag": None,
            "content_type": None,
            "content_language": None,
            "display_name": None,
        }
        == props.as_dict(raw=True)
    )

    assert props.as_dict() == {**props.raw, "type": None}
    assert props.modified is None
    assert props.created is None
    assert props.content_length is None
    assert props.content_language is None
    assert props.etag is None
    assert props.content_type is None
    assert props.display_name is None
    assert props.collection is None
    assert props.resource_type is None


def test_dav_properties():
    """Test simple dav properties extraction."""
    content = """\
    <d:response xmlns:d="DAV:">
      <d:propstat>
        <d:prop>
          <d:getlastmodified>Fri, 10 Feb 2020 10:10:10 GMT</d:getlastmodified>
          <d:creationdate>2020-1-02T03:04:05Z</d:creationdate>
          <d:getcontentlength>136314880</d:getcontentlength>
          <d:getcontentlanguage>en-US</d:getcontentlanguage>
          <d:displayname>foo</d:displayname>
          <d:resourcetype/>
          <d:getetag>"8db748065bfed5c0731e9c7ee5f9bf4c"</d:getetag>
          <d:getcontenttype>text/plain</d:getcontenttype>
        </d:prop>
      </d:propstat>
    </d:response>"""
    elem = fromstring(content)

    props = DAVProperties(elem)
    assert props.response_xml == elem
    assert (
        props.raw
        == {
            "modified": "Fri, 10 Feb 2020 10:10:10 GMT",
            "created": "2020-1-02T03:04:05Z",
            "content_length": "136314880",
            "etag": '"8db748065bfed5c0731e9c7ee5f9bf4c"',
            "content_type": "text/plain",
            "content_language": "en-US",
            "display_name": "foo",
        }
        == props.as_dict(raw=True)
    )
    assert props.created == datetime(2020, 1, 2, 3, 4, 5, tzinfo=tzutc())
    assert props.modified == datetime(2020, 2, 10, 10, 10, 10, tzinfo=tzutc())
    assert props.content_length == 136314880
    assert props.etag == '"8db748065bfed5c0731e9c7ee5f9bf4c"'
    assert props.content_type == "text/plain"
    assert props.content_language == "en-US"
    assert props.display_name == "foo"
    assert props.collection is False
    assert props.resource_type == "file"

    assert props.as_dict() == {
        "created": datetime(2020, 1, 2, 3, 4, 5, tzinfo=tzutc()),
        "modified": datetime(2020, 2, 10, 10, 10, 10, tzinfo=tzutc()),
        "content_length": 136314880,
        "etag": '"8db748065bfed5c0731e9c7ee5f9bf4c"',
        "content_type": "text/plain",
        "content_language": "en-US",
        "display_name": "foo",
        "type": "file",
    }


def test_dav_properties_partial():
    """Test that it is still readable if there's partial propfind response."""
    content = """\
    <d:propstat xmlns:d="DAV:">
      <d:prop>
        <d:displayname>foo</d:displayname>
        <d:resourcetype><d:collection/></d:resourcetype>
      </d:prop>
    </d:propstat>"""
    elem = fromstring(content)

    props = DAVProperties(elem)
    assert (
        props.raw
        == {
            "modified": None,
            "created": None,
            "content_length": None,
            "etag": None,
            "content_type": None,
            "content_language": None,
            "display_name": "foo",
        }
        == props.as_dict(raw=True)
    )

    assert props.as_dict() == {**props.raw, "type": "directory"}
    assert props.modified is None
    assert props.created is None
    assert props.content_length is None
    assert props.content_language is None
    assert props.etag is None
    assert props.content_type is None
    assert props.display_name == "foo"
    assert props.collection is True
    assert props.resource_type == "directory"


@pytest.mark.parametrize(
    "href, absolute",
    [
        ("/remote.php/dav/files/admin/dir/", False),
        ("https://example.org/remote.php/dav/files/admin/dir/", True),
    ],
)
def test_response(href: str, absolute: bool):
    """Test common scenarios for Response object."""
    content = f"""
    <d:response xmlns:d="DAV:">
        <d:href>{href}</d:href>
        <d:status>HTTP/1.1 423 Locked</d:status>
        <d:error><d:lock-token-submitted/></d:error>
    </d:response>
    """
    elem = fromstring(content)

    response = Response(elem)
    assert response.response_xml == elem
    assert response.href == href
    assert response.is_href_absolute == absolute
    assert response.path == "/remote.php/dav/files/admin/dir/"
    assert response.path_norm == "/remote.php/dav/files/admin/dir"
    assert response.error == ""
    assert response.has_propstat is False

    assert isinstance(response.properties, DAVProperties)
    assert response.response_description is None
    assert response.properties.response_xml is elem
    assert response.status_code == 423
    assert response.reason_phrase == "Locked"
    assert response.location is None

    assert str(response) == f"Response: {response.path_norm}"
    assert repr(response) == f"Response: {response.path}"


@pytest.mark.parametrize(
    "href",
    [
        "/remote.php/dav/files/admin/dir/",
        "/remote.php/dav/files/admin/dir",
        "https://example.org/remote.php/dav/files/admin/dir/",
        "https://example.org/remote.php/dav/files/admin/dir",
    ],
)
@pytest.mark.parametrize(
    "base_url, expected",
    [
        ("https://example.org", "remote.php/dav/files/admin/dir"),
        ("https://example.org/", "remote.php/dav/files/admin/dir"),
        ("https://example.org/remote.php", "dav/files/admin/dir"),
        ("https://example.org/remote.php/", "dav/files/admin/dir"),
        ("https://example.org/remote.php/dav", "files/admin/dir"),
        ("https://example.org/remote.php/dav/", "files/admin/dir"),
        ("https://example.org/remote.php/dav/files/admin/dir", "/"),
        ("https://example.org/remote.php/dav/files/admin/dir", "/"),
    ],
)
def test_path_relative_to(href: str, base_url: str, expected: str):
    """Test path relative to calculation from Response of the resource."""
    content = f"""
    <d:response xmlns:d="DAV:">
        <d:href>{href}</d:href>
        <d:status>HTTP/1.1 423 Locked</d:status>
        <d:error><d:lock-token-submitted/></d:error>
    </d:response>
    """
    elem = fromstring(content)

    response = Response(elem)
    assert response.href == href
    assert response.path_relative_to(URL(base_url)) == expected


def test_get_response_for_path():
    """Test getting response for appropriate path."""
    content = """
    <d:multistatus xmlns:d="DAV:">
        <d:response>
            <d:href>/remote.php/dav/files/admin/sample1.txt</d:href>
            <d:propstat>
                <d:prop><d:resourcetype/></d:prop>
                <d:status>HTTP/1.1 200 OK</d:status>
            </d:propstat>
        </d:response>
        <d:response>
            <d:href>/remote.php/dav/files/admin/sample2.txt</d:href>
            <d:propstat>
                <d:prop><d:resourcetype><d:collection/></d:resourcetype></d:prop>
                <d:status>HTTP/1.1 200 OK</d:status>
            </d:propstat>
        </d:response>
    </d:multistatus>
    """

    response = MultiStatusResponse(content)
    response.raise_for_status()
    assert response.content == content

    expected_keys = [
        "/remote.php/dav/files/admin/sample1.txt",
        "/remote.php/dav/files/admin/sample2.txt",
    ]

    assert list(response.responses.keys()) == expected_keys
    assert response.response_description is None

    prop1 = response.responses[expected_keys[0]].properties
    assert prop1.resource_type == "file"

    prop2 = response.responses[expected_keys[1]].properties
    assert prop2.resource_type == "directory"

    assert (
        response.get_response_for_path(
            "/remote.php/dav/files", "admin/sample1.txt"
        )
        == response.responses[expected_keys[0]]
    )
    assert (
        response.get_response_for_path(
            "/remote.php/dav/files/admin", "sample2.txt"
        )
        == response.responses[expected_keys[1]]
    )


def test_raise_for_status():
    """Test raising exception from multistatus response."""
    content = """\
    <d:multistatus xmlns:d='DAV:'>
        <d:response>
            <d:href>/othercontainer/C2/</d:href>
            <d:status>HTTP/1.1 423 Locked</d:status>
            <d:error><d:lock-token-submitted/></d:error>
        </d:response>
        <d:response>
            <d:href>/othercontainer/C3/</d:href>
            <d:status>HTTP/1.1 423 Locked</d:status>
            <d:error><d:lock-token-submitted/></d:error>
        </d:response>
    </d:multistatus>
    """
    response = MultiStatusResponse(content)
    with pytest.raises(MultiStatusResponseError) as exc_info:
        response.raise_for_status()

    expected = {
        "/othercontainer/C2/": "Locked",
        "/othercontainer/C3/": "Locked",
    }
    assert str(exc_info.value) == f"multiple errors received: {str(expected)}"
    assert exc_info.value.statuses == expected


def test_prepare_propfind_data_empty():
    """Test preparing propfind data, without specifying any parameters.

    Should return empty in this case.
    """
    return prepare_propfind_request_data() is None


@pytest.mark.parametrize(
    ["name", "namespace", "expected_inner_element"],
    [
        ("created", None, '<prop><ns0:creationdate xmlns="" />'),
        ("modified", "newDAV", '<prop><ns0:getlastmodified xmlns="newDAV" />'),
        ("something", None, '<prop><ns0:something xmlns="" />'),
    ],
    # unmapped as in, we haven't mapped this property internally
    # as we do with `created -> creationdate` and `modified -> getlastmodified`
    # it should be sent as-is.
    ids=["simple", "namespaced", "unmapped_property"],
)
def test_prepare_propfind_data(
    name: str, namespace: Optional[str], expected_inner_element: str
):
    """Test preparing propfind xml string data."""
    body = prepare_propfind_request_data(name, namespace=namespace)
    expected = '<propfind xmlns:ns0="DAV:" xmlns="DAV:">{0}</prop></propfind>'
    assert body == expected.format(expected_inner_element)


def test_try_parse_multistatus_response_for_not_a_207_response():
    """Test trying to parse with a multistatus response that's not 207."""
    with pytest.raises(ValueError) as exc_info:
        parse_multistatus_response(HTTPResponse(status_code=404))

    assert str(exc_info.value) == "http response is not a multistatus response"


def test_parse_multistatus_response():
    """Test trying to parse multistatus response."""
    res = parse_multistatus_response(
        HTTPResponse(status_code=207, text="<d></d>")
    )
    assert isinstance(res, MultiStatusResponse)
    assert res.responses == {}
    assert res.response_description is None
    assert res.content == "<d></d>"

    res.raise_for_status()


def test_check_multistatus():
    """Test MultiStatusError string representation."""
    statuses = {"/data/bar": "Locked"}
    error = MultiStatusResponseError(statuses)
    assert str(error) == "The resource /data/bar is locked"

    statuses = {"/data/bar": "Locked", "http://example.org": "Bad Gateway"}
    error = MultiStatusResponseError(statuses)
    assert str(error) == "multiple errors received: " + str(statuses)



"""Testing url parsing and processing logic."""

import pytest

from webdav4.urls import (
    URL,
    join_url,
    join_url_path,
    normalize_path,
    relative_url_to,
    strip_leading_slash,
)


@pytest.mark.parametrize(
    "base_url, path, expected",
    [
        ("https://example.org", "/", "https://example.org/"),
        ("https://example.org/", "/", "https://example.org/"),
        ("https://example.org", "", "https://example.org/"),
        ("https://example.org/", "", "https://example.org/"),
        ("https://example.org", "path", "https://example.org/path"),
        ("https://example.org/", "path", "https://example.org/path"),
        ("https://example.org/", "/path", "https://example.org/path"),
        ("https://example.org/", "/path/", "https://example.org/path"),
        ("https://example.org/", "path/", "https://example.org/path"),
        ("https://example.org/foo", "bar", "https://example.org/foo/bar"),
        ("https://example.org/foo/", "bar", "https://example.org/foo/bar"),
        ("https://example.org/foo/", "/bar", "https://example.org/foo/bar"),
        ("https://example.org/foo", "/bar", "https://example.org/foo/bar"),
        ("https://example.org/foo/", "/bar/", "https://example.org/foo/bar"),
    ],
)
def test_join_url(base_url: str, path: str, expected: str):
    """Testing join_url operation in the client."""
    assert join_url(URL(base_url), path) == expected


@pytest.mark.parametrize(
    "base, rel, expected",
    [
        ("https://example.org/foo", "foo", "/"),
        ("https://example.org/foo", "/foo", "/"),
        ("https://example.org/foo", "foo/", "/"),
        ("https://example.org/foo", "/foo/", "/"),
        ("https://example.org/foo/", "foo", "/"),
        ("https://example.org/foo/", "/foo", "/"),
        ("https://example.org/foo/", "foo/", "/"),
        ("https://example.org/foo/", "/foo/", "/"),
        ("https://example.org", "/", "/"),
        ("https://example.org/", "/", "/"),
        ("https://example.org", "", "/"),
        ("https://example.org/", "", "/"),
        ("https://example.org", "data", "data"),
        ("https://example.org", "/data", "data"),
        ("https://example.org", "data/", "data"),
        ("https://example.org", "/data/", "data"),
        ("https://example.org/", "data", "data"),
        ("https://example.org/", "/data", "data"),
        ("https://example.org/", "data/", "data"),
        ("https://example.org/", "/data/", "data"),
        ("https://example.org/foo", "foo/bar", "bar"),
        ("https://example.org/foo", "/foo/bar", "bar"),
        ("https://example.org/foo", "foo/bar/", "bar"),
        ("https://example.org/foo", "/foo/bar/", "bar"),
        ("https://example.org/foo/", "foo/bar", "bar"),
        ("https://example.org/foo/", "/foo/bar", "bar"),
        ("https://example.org/foo/", "foo/bar/", "bar"),
        ("https://example.org/foo/", "/foo/bar/", "bar"),
    ],
)
def test_path_relative_to(base: str, rel: str, expected: str):
    """Test relative path calculation."""
    assert relative_url_to(URL(base), rel) == expected


@pytest.mark.parametrize(
    "base_path, path, expected",
    [
        ("", "", "/"),
        ("", "/", "/"),
        ("/", "", "/"),
        ("/", "/", "/"),
        ("", "foo", "/foo"),
        ("", "/foo", "/foo"),
        ("", "foo/", "/foo"),
        ("", "/foo/", "/foo"),
        ("/", "foo", "/foo"),
        ("/", "/foo", "/foo"),
        ("/", "foo/", "/foo"),
        ("/", "/foo/", "/foo"),
        ("/foo", "bar", "/foo/bar"),
        ("/foo", "/bar", "/foo/bar"),
        ("/foo", "bar/", "/foo/bar"),
        ("/foo", "/bar/", "/foo/bar"),
        ("foo/", "bar", "/foo/bar"),
        ("foo/", "/bar", "/foo/bar"),
        ("foo/", "bar/", "/foo/bar"),
        ("foo/", "/bar/", "/foo/bar"),
        ("foo", "bar", "/foo/bar"),
        ("foo", "/bar", "/foo/bar"),
        ("foo", "bar/", "/foo/bar"),
        ("foo", "/bar/", "/foo/bar"),
        ("/foo/", "bar", "/foo/bar"),
        ("/foo/", "/bar", "/foo/bar"),
        ("/foo/", "bar/", "/foo/bar"),
        ("/foo/", "/bar/", "/foo/bar"),
        ("/foo/bar", "foobar", "/foo/bar/foobar"),
        ("/foo/bar/", "foobar", "/foo/bar/foobar"),
        ("/foo/bar", "foobar/", "/foo/bar/foobar"),
        ("/foo/bar", "foobar/foobar", "/foo/bar/foobar/foobar"),
    ],
)
def test_join_url_path(base_path: str, path: str, expected: str):
    """Test joining base path and path together, while normalizing the url."""
    assert join_url_path(base_path, path) == expected


@pytest.mark.parametrize(
    "path, expected",
    [
        ("/", "/"),
        ("foo", "foo"),
        ("/foo", "/foo"),
        ("/foo/bar/", "/foo/bar"),
        ("/foo//bar//", "/foo/bar"),
        ("/////foo////bar////", "/foo/bar"),
    ],
)
def test_normalize_url(path: str, expected: str):
    """Test that it normalizes urls removing too many "/" and leading slash."""
    assert normalize_path(path) == expected


@pytest.mark.parametrize(
    "path, expected",
    [
        ("", ""),
        ("/", "/"),
        ("foo", "foo"),
        ("foo/bar/", "foo/bar"),
        ("/foo/bar/", "/foo/bar"),
    ],
)
def test_strip_leading_slash(path: str, expected: str):
    """Test that it strips leading slash, except the "/" only urls."""
    assert strip_leading_slash(path) == expected



"""Testing http client."""

from unittest import mock

from webdav4.http import Client


def test_webdav_methods():
    """Test webdav methods that are added on httpx.Client."""
    url = "https://example.org"
    client = Client(auth=("user", "password"))
    response = mock.Mock()

    with mock.patch.object(client, "request", return_value=response) as m:
        assert client.propfind(url, data="<xml/>") == response
        m.assert_called_once_with("PROPFIND", url, data="<xml/>")

    with mock.patch.object(client, "request", return_value=response) as m:
        assert client.proppatch(url, data="<xml/>") == response
        m.assert_called_once_with("PROPPATCH", url, data="<xml/>")

    with mock.patch.object(client, "request", return_value=response) as m:
        assert client.mkcol(url) == response
        m.assert_called_once_with("MKCOL", url)

    headers = {"Destination": "https://example.com"}

    with mock.patch.object(client, "request", return_value=response) as m:
        assert client.copy(url, headers=headers) == response
        m.assert_called_once_with("COPY", url, headers=headers)

    with mock.patch.object(client, "request", return_value=response) as m:
        assert client.move(url, headers=headers) == response
        m.assert_called_once_with("MOVE", url, headers=headers)

    with mock.patch.object(client, "request", return_value=response) as m:
        assert client.lock(url) == response
        m.assert_called_once_with("LOCK", url)

    with mock.patch.object(client, "request", return_value=response) as m:
        assert client.unlock(url) == response
        m.assert_called_once_with("UNLOCK", url)



"""Test fs utils."""
from io import BytesIO

from webdav4.fs_utils import peek_filelike_length

from .test_callback import ReadWrapper


def test_peek_filelike_length():
    """Test peek_filelike length for the fileobj."""
    fobj = BytesIO(b"Hello, World!")

    assert peek_filelike_length(fobj) == 13
    assert peek_filelike_length(ReadWrapper(fobj)) is None  # type: ignore
    assert peek_filelike_length(object()) is None  # type: ignore



"""Test IO callback wrapper."""
from collections.abc import Iterable
from io import StringIO
from pathlib import Path
from typing import Union, no_type_check
from unittest.mock import MagicMock, call

import pytest

from webdav4.callback import CallbackIOWrapper, wrap_file_like


def test_wrap_file_like():
    """Test wrapper."""
    buff = StringIO()
    callback = MagicMock()
    result = wrap_file_like(buff, callback)
    assert isinstance(result, CallbackIOWrapper)


@no_type_check
def test_callback_read():
    """Test reports read callback correctly."""
    line = "foo\n"
    buff = StringIO(line * 100)
    buff.seek(0)

    callback = MagicMock()

    wrapper = CallbackIOWrapper(buff, callback)
    assert isinstance(wrapper, Iterable)

    assert wrapper.read(20) == line * 5
    callback.assert_called_once_with(20)
    callback.reset_mock()

    assert wrapper.read(40) == line * 10
    callback.assert_called_once_with(40)
    callback.reset_mock()

    # getattr works
    assert wrapper.closed is False

    # adding to __dict__ works, and is being wrapped properly
    assert wrapper.__wrapped_stream__ is buff

    method = wrapper.write
    assert method.__self__ is buff and method == buff.write
    assert wrapper.read != buff.read
    assert wrapper.__call_back__ is callback

    # setattr works
    wrapper.add_something_here = True
    assert wrapper.add_something_here
    assert buff.add_something_here


@no_type_check
def test_callback_write():
    """Test reports write callback correctly."""
    line = "foo\n"
    read_buffer = StringIO(line * 100)
    read_buffer.seek(0)

    write_buffer = StringIO()
    callback = MagicMock()

    wrapper = CallbackIOWrapper(write_buffer, callback, method="write")
    assert isinstance(wrapper, Iterable)

    # write works
    assert wrapper.write(read_buffer.read(20)) == 20
    callback.assert_called_once_with(20)
    callback.reset_mock()

    assert wrapper.write(read_buffer.read(40)) == 40
    callback.assert_called_once_with(40)
    callback.reset_mock()

    # getattr works
    assert wrapper.closed is False

    # __dict__ overwrite works, wrapped function gets called
    assert wrapper.__wrapped_stream__ is write_buffer

    method = wrapper.read
    assert method.__self__ is write_buffer and method == write_buffer.read
    assert wrapper.write != write_buffer.write

    assert wrapper.__call_back__ is callback

    # setattr works
    wrapper.add_something_here = True
    assert wrapper.add_something_here
    assert write_buffer.add_something_here


class ReadWrapper:
    """Wraps any given buffer."""

    def __init__(self, buff):
        """Wrap buff."""
        self.buff = buff

    def read(self, *args):
        """Wrap read method of the buff."""
        return self.buff.read(*args)


@pytest.mark.parametrize("mode", ["r", "rb"])
@pytest.mark.parametrize("no_iter_implemented", [True, False])
@no_type_check
def test_callback_read_iter(
    tmp_path: Path, mode: str, no_iter_implemented: bool
):
    """Test __iter__ callbacks."""
    path = tmp_path / "file.txt"

    with path.open(mode="w", newline="\n", encoding="utf-8") as f:
        for _ in range(100):
            f.write("foo\n")

    callback = MagicMock()

    def decode(ch: Union[str, bytes]) -> Union[str, bytes]:
        return ch.decode("utf-8") if isinstance(ch, bytes) else ch

    with path.open(mode=mode) as f:
        wrapper = CallbackIOWrapper(
            ReadWrapper(f) if no_iter_implemented else f, callback
        )

        chunks = list(wrapper)
        assert ["foo\n" for _ in chunks] == list(map(decode, chunks))

    callback.assert_has_calls([call(4) for _ in chunks])


def test_callback_iter_non_read_method():
    """Error if __iter__() not implemented and we are in non-read mode."""

    class ReadWriteWrapper(ReadWrapper):
        """Wraps both read and write method of the buff."""

        def write(self, *args):
            """Wrap write method of the buff."""
            return self.buff.write(*args)  # pragma: no cover

    callback = MagicMock()

    f = ReadWriteWrapper(StringIO())  # type: ignore
    with pytest.raises(TypeError) as exc_info:
        wrapper = CallbackIOWrapper(
            f, callback, method="write"  # type: ignore
        )
        for _ in wrapper:
            pass
    assert str(exc_info.value) == "'ReadWriteWrapper' object is not iterable"


def test_callback_illegal_method():
    """Test that other methods except read/write are unsupported."""
    f = StringIO()
    callback = MagicMock()
    with pytest.raises(ValueError) as exc_info:
        CallbackIOWrapper(f, callback, method="unknown")  # type: ignore

    assert str(exc_info.value) == "Can only wrap read/write methods"
